relativePath = "./src/data/"


def readData(fileName: str, array):
    touch(relativePath + fileName)
    file = open(relativePath + fileName, "r")
    lines = file.readlines()
    file.close()

    readArray = [
        [n for n in l.split(",")] for l in lines
    ]

    rows = range(min(len(readArray), len(array)))

    for rowNum in rows:
        for colNum in rows:
            readVal = readArray[rowNum][colNum]
            if (readVal == "None" or readVal == "None\n"):
                readVal = None
            else:
                readVal = float(readVal)
            array[rowNum][colNum] = readVal


def touch(fileName: str):
    file = open(fileName, "a+")
    file.close()


def writeData(fileName: str, array):
    file = open(relativePath + fileName, "w")
    for row in array:
        file.write(",".join([str(x) for x in row])+"\n")
    file.close()



from riskCalculator.probabilityOfLosses import *


def recursiveMarkovChain(attackers: int, defenders: int, memoryArray, baseCase):
    if (defenders <= 0 or attackers <= 0):
        return baseCase(attackers)

    rememberedValue = memoryArray[attackers-1][defenders-1]
    if (rememberedValue != None):
        return rememberedValue

    attackDie = min(attackers, 3)
    defenseDie = min(defenders, 2)

    # expected value of the chain(a,d) = \sum_l P(l)*chain(a-l, d - dd + l)
    result = sum(
        [
            # the probability of losses
            probabilityOfLosses(attackDie, defenseDie, losses)
            # the expected value of the scenario after losses
            * recursiveMarkovChain(attackers - losses, defenders - defenseDie + losses, memoryArray, baseCase)
            for losses in range(min(attackDie, defenseDie) + 1)
        ]
    )

    memoryArray[attackers-1][defenders-1] = result
    return result



from riskCalculator.recursiveMarkovChain import *


def expectedSurvivors(attackers: int, defenders: int, expectedSurvivorsArray):
    return recursiveMarkovChain(
        attackers,
        defenders,
        expectedSurvivorsArray,
        # if there are no defenders then all attackers survive
        lambda attackers: attackers if attackers > 0 else 0
    )



from argparse import *


def parseArguments():
    parser = ArgumentParser(description="arguments for risk calculation")
    parser.add_argument('attackers', type=int, help='number of attackers')
    parser.add_argument('defenders', type=int, help='number of defenders')
    parser.add_argument(
        '--read',
        action=BooleanOptionalAction,
        dest="read",
        help="should use saved results"
    )
    parser.add_argument(
        '--write',
        action=BooleanOptionalAction,
        dest="write",
        help="should save results"
    )
    parser.add_argument(
        '--recursion-limit',
        type=int,
        default=1000,
        metavar="R",
        dest="recursionLimit",
        help="recursion limit, default 1000"
    )

    return parser.parse_args()



import sys
from riskCalculator.parseArguments import *
from riskCalculator.readWriteData import *
from riskCalculator.likelihoodOfVictory import *
from riskCalculator.expectedSurvivors import *


def risk():
    args = parseArguments()
    attackers = args.attackers
    defenders = args.defenders

    # initialise variables
    maxDice = max(attackers, defenders, 200 if args.write else 1)
    expectedSurvivorsArray = [[None]*maxDice for x in range(maxDice)]
    likelihoodOfVictoryArray = [[None]*maxDice for x in range(maxDice)]
    sys.setrecursionlimit(args.recursionLimit)

    if (args.read):
        readData("expectedSurvivors.txt", expectedSurvivorsArray)
        readData("likelihoodOfVictory.txt", likelihoodOfVictoryArray)

    # calculate expected result and chance
    chance = round(
        likelihoodOfVictory(
            attackers, defenders, likelihoodOfVictoryArray
        ) * 100,
        1
    )

    survivors = round(
        expectedSurvivors(
            attackers, defenders, expectedSurvivorsArray
        ),
        2
    )

    # inform user
    print(
        "likelihood of victory " + str(chance) + "% with "
        + str(survivors) + " survivors on average"
    )

    if (args.write):
        writeData("expectedSurvivors.txt", expectedSurvivorsArray)
        writeData("likelihoodOfVictory.txt", likelihoodOfVictoryArray)

    return [chance, survivors]



from riskCalculator.recursiveMarkovChain import *


def likelihoodOfVictory(attackers: int, defenders: int, likelihoodOfVictoryArray):
    return recursiveMarkovChain(
        attackers,
        defenders,
        likelihoodOfVictoryArray,
        # if there are no defenders and any attackers they win with probability 1
        lambda attackers: 1 if attackers > 0 else 0
    )



def probabilityOfLosses(attackDie: int, defenseDie: int, losses: int):
    if (defenseDie == 0):
        return 0
    if (defenseDie == 1):
        return probabilityOfXLossesOneDefender(attackDie, losses)
    return probabilityOfXLossesTwoDefenders(attackDie, losses)


def probabilityOfXLossesOneDefender(attackDie: int, losses: int):
    if (losses == 0):
        return [0.417, 0.579, 0.66][attackDie - 1]
    return [0.583, 0.421, 0.34][attackDie - 1]


def probabilityOfXLossesTwoDefenders(attackDie: int, losses: int):
    if (losses == 0):
        return [0.255, 0.228, 0.372][attackDie - 1]
    if (losses == 1):
        return [0, 0.448, 0.292][attackDie - 1]
    return [0.745, 0.448, 0.336][attackDie - 1]
